import zlib
from datetime import datetime

from sqlalchemy import (Boolean, Column, DateTime, ForeignKey, Integer, String,
                        event)
from sqlalchemy.orm import relationship
from sqlalchemy.orm.session import Session

from ..chunk import CHUNKSIZE, delete_chunk, load_chunk, save_chunk
from . import Base


class FileChunk(Base):
    __tablename__ = 'file_chunks'

    id = Column(Integer, primary_key=True)
    file_id = Column(Integer, ForeignKey('files.id'))
    index = Column(Integer)
    size = Column(Integer)
    chunk_hash = Column(String)
    compressed = Column(Boolean, default=False)

    file = relationship("File", back_populates="chunks")

    @property
    def chunk(self):
        if hasattr(self, '_chunk'):
            return self._chunk
        return load_chunk(self.chunk_hash, self.compressed)

    @chunk.setter
    def chunk(self, data):
        self._chunk = data


@event.listens_for(FileChunk, 'before_insert')
def before_insert_file_chunk(mapper, connection, target: FileChunk):
    target.chunk_hash, target.size = save_chunk(target._chunk)


class File(Base):
    __tablename__ = 'files'

    id = Column(Integer, primary_key=True)
    ctime = Column(DateTime, default=datetime.utcnow)
    mtime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)
    name = Column(String)
    size = Column(Integer)
    chunks = relationship("FileChunk", order_by=FileChunk.index)

    def write(self, data):
        self.mtime = datetime.utcnow()
        self.size = len(data)
        self.chunks = []
        for i in range(0, len(data), CHUNKSIZE):
            chunk = FileChunk()
            chunk.index = i // CHUNKSIZE
            chunk._chunk = data[i:i + CHUNKSIZE]
            chunk.size = len(chunk._chunk)
            self.chunks.append(chunk)

    def read(self):
        self.atime = datetime.utcnow()
        return b''.join([c.chunk for c in self.chunks])


@event.listens_for(File, 'before_delete')
def before_delete_file(mapper, connection, target):
    for chunk in target.chunks:
        connection.execute(
            FileChunk.__table__.delete().where(FileChunk.id == chunk.id))


def compress_chunks(db: Session):
    for chunk in db.query(FileChunk).filter(
            FileChunk.compressed == False).limit(100):
        old_chunk_hash = chunk.chunk_hash
        buf = zlib.compress(chunk.chunk)
        chunk_hash, size = save_chunk(buf, compressed=True)
        chunk.chunk_hash = chunk_hash
        chunk.size = size
        chunk.compressed = True
        try:
            db.commit()
            delete_chunk(old_chunk_hash)
        except:
            db.rollback()
            delete_chunk(chunk_hash)



import itertools
import re
from typing import Callable

from .config import style
from .qt_compat import AlignRight, QtWidgets


def matched_xy_pairs(patterns: str, lst: list[str]) -> list[tuple[str, str]]:
    patterns = patterns.replace(" ", "").split(";")
    pairs = []
    for x, y in itertools.product(lst, repeat=2):
        test = f"{x},{y}"
        for pattern in patterns:
            r = re.match(pattern, test)
            if r and r.group(0) == test:
                pairs.append((x, y))
                break
    return pairs


class FormatCombo(QtWidgets.QComboBox):

    def __init__(self):
        super().__init__()
        self.on_change_callable = None

    def set_on_change_event_action(self, callback: Callable[[], None]):
        self.on_change_callable = callback
        self.activated.connect(callback)

    def set_idx(self, idx: int):
        self.setCurrentIndex(idx)
        if (callable(self.on_change_callable)):
            self.on_change_callable()


class XFormatCombo(FormatCombo):

    def __init__(self):
        super().__init__()
        self.addItem("real")
        self.addItem("imag")
        self.addItem("mag")
        self.addItem("phase")


class YFormatCombo(FormatCombo):

    def __init__(self):
        super().__init__()
        self.addItem("mag")
        self.addItem("phase")
        self.addItem("real")
        self.addItem("imag")


class LineEdit(QtWidgets.QLineEdit):

    def set_on_change_event_action(self, callback: Callable[[], None]):
        self.on_change_callable = callback
        self.editingFinished.connect(callback)

    def set_text(self, w):
        self.setText(w)
        if (callable(self.on_change_callable)):
            self.on_change_callable()


class SelectionBundle():

    def __init__(self):
        self.stx = LineEdit()
        # select text
        self.fx = XFormatCombo()
        self.fy = YFormatCombo()
        self.lx = QtWidgets.QCheckBox("logX")
        self.ly = QtWidgets.QCheckBox("logY")
        self.linkx = QtWidgets.QCheckBox("ShareX")
        self.linky = QtWidgets.QCheckBox("ShareY")
        self.sels = []
        # tuple enumeration

    def set_on_change_event_actions(self, on_text_edited, on_format_changed,
                                    on_log_scale_marker_changed):
        self.stx.set_on_change_event_action(on_text_edited)
        self.fx.set_on_change_event_action(on_format_changed)
        self.fy.set_on_change_event_action(on_format_changed)
        self.lx.toggled.connect(on_format_changed)
        self.ly.toggled.connect(on_format_changed)
        self.linkx.toggled.connect(on_log_scale_marker_changed)
        self.linky.toggled.connect(on_log_scale_marker_changed)

    def rm4l(self):  # remove from layout
        self.stx.setParent(None)
        self.fx.setParent(None)
        self.fy.setParent(None)
        self.lx.setParent(None)
        self.ly.setParent(None)
        self.linkx.setParent(None)
        self.linky.setParent(None)

    def a2l(self, layout):  # add to layout
        i = 3
        layout.addWidget(self.stx, 0, i)
        i += 2
        layout.addWidget(self.fx, 0, i)
        i += 2
        layout.addWidget(self.fy, 0, i)
        i += 2
        layout.addWidget(self.lx, 0, 10)
        layout.addWidget(self.ly, 0, 11)
        layout.addWidget(self.linkx, 0, 12)
        layout.addWidget(self.linky, 0, 13)


class ToolBar(QtWidgets.QWidget):

    def __init__(self):
        super().__init__()

        # the buttons
        # button for points
        self.mode = 'P'
        self.setStyleSheet(style)

        self.pb = QtWidgets.QRadioButton('Points')
        self.pb.setChecked(True)
        self.pb.toggled.connect(self.toggle_mode)

        # button for Traces
        self.tb = QtWidgets.QRadioButton('Traces')
        self.tb.toggled.connect(self.toggle_mode)

        # text labels
        self.ytxt_lb = QtWidgets.QLabel("(X,Y)")
        self.ytxt_lb.setAlignment(AlignRight)
        self.fx_lb = QtWidgets.QLabel("fx")
        self.fx_lb.setAlignment(AlignRight)
        self.fy_lb = QtWidgets.QLabel("fy")
        self.fy_lb.setAlignment(AlignRight)

        # enumeration
        ps = SelectionBundle()
        ps.set_on_change_event_actions(self.textEdited, self.generateXYFM,
                                       self.link_edited)

        ts = SelectionBundle()
        ts.set_on_change_event_actions(self.textEdited, self.generateXYFM,
                                       self.link_edited)

        # connections :
        self.ps = ps
        self.ts = ts

        # plot format configures
        self.xypairs = []
        self.xypairs_dirty = True
        self.fx = None
        self.fy = None
        self.xyfm_dirty = True
        self.link_dirty = True
        self.lx = False
        self.ly = False
        # setting layout
        self.layout = QtWidgets.QGridLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        self.layout.addWidget(self.pb, 0, 0)
        self.layout.addWidget(self.tb, 0, 1)

        self.layout.addWidget(self.ytxt_lb, 0, 2)
        self.layout.addWidget(self.fx_lb, 0, 4)
        self.layout.addWidget(self.fy_lb, 0, 6)

        self.AR = QtWidgets.QPushButton("AR")
        self.AR.setMaximumWidth(30)
        self.AR.setToolTip("Auto Range")
        self.CR = QtWidgets.QPushButton("CLR")
        self.CR.setMaximumWidth(30)
        self.CR.setToolTip("Clearing History Plots")
        self.CR_flag = False
        self.layout.addWidget(self.AR, 0, 8)
        self.layout.addWidget(self.CR, 0, 9)

        self.refresh_layout()

    @property
    def column_names(self) -> list[str]:
        return {
            "P": self.mainwindow.point_data_box.column_names,
            "T": self.mainwindow.trace_data_box.column_names
        }[self.mode]

    @property
    def selections(self) -> SelectionBundle:
        return {"P": self.ps, "T": self.ts}[self.mode]

    def set_trace_text(self, text: str):
        self.ts.stx.set_text(text)

    def set_point_text(self, text: str):
        self.ps.stx.set_text(text)

    def sharexy(self):
        return self.selections.linkx.isChecked(
        ), self.selections.linky.isChecked()

    def set_mainwindow(self, mainwindow):
        self.mainwindow = mainwindow
        self.AR.clicked.connect(self.mainwindow.all_enable_auto_range)
        self.CR.clicked.connect(self.CR_action)

    def AR_action(self):
        self.mainwindow.enable_all_auto_range()

    def CR_action(self):
        self.CR_flag = True

    def refresh_layout(self):
        if self.mode == 'P':
            self.ts.rm4l()
            self.ps.a2l(self.layout)
        elif self.mode == 'T':
            self.ps.rm4l()
            self.ts.a2l(self.layout)

    def toggle_mode(self):
        if (self.pb.isChecked()):
            self.mode = 'P'
        elif (self.tb.isChecked()):
            self.mode = 'T'
        self.refresh_layout()
        self.refresh_comb()

    def refresh_comb(self, ):
        self.generateXYFM()
        self.textEdited()
        #set tooltips
        self.ytxt_lb.setToolTip(str(self.column_names))

    def link_edited(self):
        # print("LinkEdited")
        self.link_dirty = True

    def generateXYFM(self):
        self.fx = self.selections.fx.currentText()
        self.fy = self.selections.fy.currentText()
        self.lx = self.selections.lx.isChecked()
        self.ly = self.selections.ly.isChecked()
        self.xyfm_dirty = True
        #self.show_info() ;

    def textEdited(self):
        new_pairs = matched_xy_pairs(self.selections.stx.text(),
                                     self.column_names)
        if (len(self.xypairs) != len(new_pairs)):
            self.xypairs_dirty = True
        else:
            for i, xy in enumerate(new_pairs):
                if (xy != self.xypairs[i]):
                    self.xypairs_dirty = True
                    break
        self.xypairs = new_pairs



import pickle
import zlib
from typing import Any, Callable, TypeVar

import msgpack

__index = 0
__pack_handlers = {}
__unpack_handlers = {}
__compress_level = zlib.Z_NO_COMPRESSION

cls = TypeVar('cls')


def compress_level(level: int) -> None:
    """
    Set compress level

    Args:
        level: int
            An integer from 0 to 9 controlling the level of compression;
            1 is fastest and produces the least compression, 9 is slowest
            and produces the most. 0 is no compression. The default value
            is 0.
    """
    global __compress_level
    __compress_level = level


def register(cls: type,
             encode: Callable[[cls], bytes] = pickle.dumps,
             decode: Callable[[bytes], cls] = pickle.loads) -> None:
    """
    Register a serializable type

    Args:
        cls: type
        encode: Callable
            translate an object of type `cls` into `bytes`
            default: pickle.dumps
        decode: Callable
            translate `bytes` to an object of type `cls`
            default: pickle.loads
    """
    global __index
    __index += 1
    t = __index
    __pack_handlers[cls] = (t, encode)
    __unpack_handlers[t] = decode


def default(obj: Any) -> msgpack.ExtType:
    for cls, (t, encode) in __pack_handlers.items():
        if isinstance(obj, cls):
            return msgpack.ExtType(t, encode(obj))
    else:
        raise TypeError("Unknown type: %r" % (obj, ))


def ext_hook(code: int, data: bytes) -> msgpack.ExtType:
    for c, decode in __unpack_handlers.items():
        if code == c:
            return decode(data)
    else:
        return msgpack.ExtType(code, data)


def pack(obj: Any) -> bytes:
    """
    Serialize
    """
    return msgpack.packb(obj, default=default, use_bin_type=True)


def unpack(buff: bytes) -> Any:
    """
    Unserialize
    """
    return msgpack.unpackb(buff, ext_hook=ext_hook, raw=False)


def packz(obj: Any) -> bytes:
    """
    Serialize and compress.
    """
    return zlib.compress(pack(obj), level=__compress_level)


def unpackz(buff: bytes) -> Any:
    """
    Decompress and unserialize.
    """
    return unpack(zlib.decompress(buff))


def encode_excepion(e: Exception) -> bytes:
    e.__traceback__ = None
    return pickle.dumps(e)


register(Exception, encode_excepion)


try:
    import numpy as np

    dtypes = [
        np.bool_, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16,
        np.uint32, np.uint64, np.float16, np.float32, np.float64, np.complex64,
        np.complex128
    ]

    _dtype_map1 = {t: i for i, t in enumerate(dtypes)}
    _dtype_map2 = {i: t for i, t in enumerate(dtypes)}

    def encode_ndarray(x: np.ndarray) -> bytes:
        dtype = x.dtype
        if isinstance(dtype, np.dtype):
            dtype = dtype.type
        if x.flags['F_CONTIGUOUS']:
            x = x.T
            T = True
        else:
            T = False
        return pack((_dtype_map1[dtype], x.shape, T, x.data))

    def decode_ndarray(buff: bytes) -> np.ndarray:
        t, shape, T, buff = unpack(buff)
        x = np.ndarray(shape, dtype=_dtype_map2[t], buffer=buff, order='C')
        if T:
            x = x.T
        return x

    register(np.ndarray, encode_ndarray, decode_ndarray)

except:
    pass

__all__ = ['compress_level', 'register', 'pack', 'unpack', 'packz', 'unpackz']



import shutil
from abc import ABC, abstractmethod
from datetime import date, datetime
from pathlib import Path
from typing import Any
from uuid import UUID, uuid1, uuid5

from .file import load


class BaseStorage(ABC):

    @abstractmethod
    def _create_dataset(self) -> UUID:
        """Create a new dataset and return its id"""
        pass

    @abstractmethod
    def _delet_dataset(self, id: UUID):
        """Delete a dataset"""
        pass

    @abstractmethod
    def _dataset_getitem(self, id: UUID, key: str):
        """Get a dataset item"""
        pass

    @abstractmethod
    def _dataset_setitem(self, id: UUID, key: str, value: Any):
        """Set a dataset item"""
        pass

    @abstractmethod
    def _dataset_delitem(self, id: UUID, key: str):
        """Delete a dataset item"""
        pass

    @abstractmethod
    def _dataset_keys(self, id: UUID):
        """Get dataset keys"""
        pass

    @abstractmethod
    def _dataset_append(self, id: UUID, dataframe: dict[str, Any]):
        """Append a dataframe to a dataset"""
        pass

    @abstractmethod
    def _dataset_item_append(self, id: UUID, key: str, value: Any):
        """Append a value to a dataset item"""
        pass


class Storage(BaseStorage):

    def __init__(self, base: Path | str):
        if isinstance(base, str):
            base = Path(base)
        self.base = base
        self.namespace = uuid5(UUID('f89f735a-791e-5a43-9ba6-f28d58601544'),
                               base.as_posix())

    def clear(self):
        shutil.rmtree(self.base)
        self.base.mkdir(parents=True)

    def uuid(self,
             name: str | None = None,
             namespace: UUID | None = None,
             seq: int = 0) -> UUID:
        if name is None:
            name = str(uuid1())
        if namespace is None:
            return uuid5(self.namespace, f"{name}{seq}")
        else:
            return uuid5(namespace, f"{name}{seq}")

    def uuid_to_path(self, uuid: UUID) -> Path:
        return self.base / uuid.hex[:2] / uuid.hex[2:4] / uuid.hex[4:]

    def create_dataset(self, title: str, tags: list[str] = []):
        from .dataset import Dataset
        id = self.uuid()
        return Dataset(id, self)

    def get_dataset(self, id: UUID):
        from .dataset import Dataset
        return Dataset(id, self)

    def remove_dataset(self, id: UUID):
        shutil.rmtree(self.uuid_to_path(id))

    def query(self, title_pattern: str, tags: list[str],
              before: datetime | date, after: datetime | date):
        pass



from typing import Any, Sequence

import nevergrad as ng
import numpy as np
from scipy.optimize import OptimizeResult


class NgOptimizer():

    def __init__(self, variables, **kwds):
        self._all_x, self._all_y = [], []
        self.config = {
            'method': 'TBPSA',
            'budget': 100,
        }
        self.config.update(kwds)
        instrum = []
        self.dimensions = variables
        for space in variables:

            if space.transform_ == "normalize":
                instrum.append(ng.p.Scalar(init=0.5, lower=0, upper=1))
            else:
                if space.prior == "log-uniform":
                    instrum.append(
                        ng.p.Log(init=None,
                                 exponent=space.base,
                                 lower=space.low,
                                 upper=space.high))
                else:
                    instrum.append(
                        ng.p.Scalar(init=None,
                                    lower=space.low,
                                    upper=space.high))
        self.instrum = ng.p.Instrumentation(*instrum)
        self.opt = getattr(ng.optimizers,
                           self.config['method'])(self.instrum,
                                                  budget=self.config['budget'])

    def ask(self):
        tmp = self.opt.ask()
        return [
            space.inverse_transform(x)
            for x, space in zip(tmp.args, self.dimensions)
        ]

    def tell(self, suggested: Sequence, value: Any):
        self._all_x.append(suggested)
        self._all_y.append(value)
        suggested = [
            space.transform(x) for x, space in zip(suggested, self.dimensions)
        ]
        self.opt.suggest(*suggested)
        x = self.opt.ask()
        self.opt.tell(x, value)

    def get_result(self, history: bool = False) -> OptimizeResult:
        recommendation = self.opt.recommend()
        ret = OptimizeResult({
            'x': [
                space.inverse_transform(x)
                for x, space in zip(recommendation.args, self.dimensions)
            ]
        })
        if history:
            ret.x_iters = self._all_x
            ret.func_vals = self._all_y
        ret.fun = recommendation.value
        return ret



from ..net.kcp import listen


class Peer:

    def __init__(self, addr, handler, **kwargs):
        self.addr = addr
        self.handler = handler
        self.kwargs = kwargs

    async def run(self):
        conv = self.kwargs.get('conv', 0)
        ttl = self.kwargs.get('ttl', 60)
        async with listen(self.addr, conv, ttl, **self.kwargs) as server:
            async for conn in server:
                await self.handler(conn)

    def __call__(self, handler):
        self.handler = handler
        return self

    def __await__(self):
        return self.run().__await__()

    def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass
